@@ -1,7 +1,7 @@
 /*
-    Copyright (c)  2006		    Dmitry K. Butskoy
-				    <buc@citadel.stu.neva.ru>
-    License:  GPL		
+    Copyright (c)  2006, 2007		Dmitry Butskoy
+					<buc@citadel.stu.neva.ru>
+    License:  GPL v2 or any later
 
     See COPYING for the status of this software.
 */
@@ -11,20 +11,16 @@
 #include <stdarg.h>
 #include <unistd.h>
 #include <fcntl.h>
-#include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/poll.h>
-#include <time.h>
-#include <sys/time.h>
 #include <netinet/icmp6.h>
 #include <netinet/ip_icmp.h>
 #include <netinet/in.h>
 #include <netinet/ip6.h>
-#include <arpa/inet.h>
 #include <netdb.h>
-#include <net/if.h>
 #include <errno.h>
 #include <locale.h>
+#include <sys/utsname.h>
 #include <linux/types.h>
 #include <linux/errqueue.h>
 
@@ -63,15 +59,12 @@
 #define DEF_AF		AF_INET
 
 
-#define __TEXT(X)       #X
-#define _TEXT(X)        __TEXT(X)
-
 static char version_string[] = "Modern traceroute for Linux, "
 				"version " _TEXT(VERSION) ", " __DATE__
 				"\nCopyright (c) 2006  Dmitry Butskoy, "
 				"  License: GPL";
 static int debug = 0;
-static unsigned int first_hop = 0;
+static unsigned int first_hop = 1;
 static unsigned int max_hops = DEF_HOPS;
 static unsigned int sim_probes = DEF_SIM_PROBES;
 static unsigned int probes_per_hop = DEF_NUM_PROBES;
@@ -96,8 +89,14 @@ static sockaddr_any dst_addr = {{ 0, }, };
 static char *dst_name = NULL;
 static char *device = NULL;
 static sockaddr_any src_addr = {{ 0, }, };
+static unsigned int src_port = 0;
+
+static const char *module = "default";
+static const tr_module *ops = NULL;
+
+static char *opts[16] = { NULL, };	/*  assume enough   */
+static unsigned int opts_idx = 1;	/*  first one reserved...   */
 
-static tr_ops *ops = &udp_ops;
 
 static int af = 0;
 
@@ -145,9 +144,9 @@ static void check_progname (const char *name) {
 	else if (p[l] == '4')  af = AF_INET;
 
 	if (!strncmp (p, "tcp", 3))
-		ops = &tcp_ops;
+		module = "tcp";
 	if (!strncmp (p, "tracert", 7))
-		ops = &icmp_ops;
+		module = "icmp";
 
 	return;
 }
@@ -218,6 +217,7 @@ static const char *addr2str (const sockaddr_any *addr) {
 
 
 /*	IP  options  stuff	    */
+
 static void init_ip_options (void) {
 	sockaddr_any *gates;
 	int i, max;
@@ -326,14 +326,66 @@ static int set_source (CLIF_option *optn, char *arg) {
 	return  getaddr (arg, &src_addr);
 }
 
-	
-static int set_ops (CLIF_option *optn, char *arg) {
+static int set_port (CLIF_option *optn, char *arg) {
+	unsigned int *up = (unsigned int *) optn->data;
+	char *q;
 
-	ops = (tr_ops *) optn->data;
+	*up = strtoul (arg, &q, 0);
+	if (q == arg) {
+	    struct servent *s = getservbyname (arg, NULL);
+
+	    if (!s)  return -1;
+	    *up = ntohs (s->s_port);
+	}
 
 	return 0;
 }
 
+static int set_module (CLIF_option *optn, char *arg) {
+
+	module = (char *) optn->data;
+
+	return 0;
+}
+
+
+static int set_mod_option (CLIF_option *optn, char *arg) {
+
+	if (!strcmp (arg, "help")) {
+	    const tr_module *mod = tr_get_module (module);
+
+	    if (mod && mod->options) {
+		/*  just to set common keyword flag...  */
+		CLIF_parse (1, &arg, 0, 0, CLIF_KEYWORD);
+		CLIF_print_options (NULL, mod->options);
+	    } else
+		fprintf (stderr, "No options for module `%s'\n", module);
+
+	    exit (0);
+	}
+
+	if (opts_idx >= sizeof (opts) / sizeof (*opts))  {
+	    fprintf (stderr, "Too many module options\n");
+	    return -1;
+	}
+
+	opts[opts_idx] = strdup (arg);
+	if (!opts[opts_idx])  error ("strdup");
+	opts_idx++;
+
+	return 0;
+}
+
+
+static int set_raw (CLIF_option *optn, char *arg) {
+	char buf[1024];
+
+	module = "raw";
+
+	snprintf (buf, sizeof (buf), "protocol=%s", arg);
+	return  set_mod_option (optn, buf);
+}
+
 
 static int set_host (CLIF_argument *argm, char *arg, int index) {
 
@@ -363,11 +415,9 @@ static CLIF_option option_list[] = {
 			    _TEXT(MAX_GATEWAYS_6) " for IPv6)",
 			add_gateway, 0, 0, CLIF_SEVERAL },
 	{ "I", "icmp", 0, "Use ICMP ECHO for tracerouting",
-			set_ops, &icmp_ops, 0, 0 },
+			set_module, "icmp", 0, 0 },
 	{ "T", "tcp", 0, "Use TCP SYN for tracerouting",
-			set_ops, &tcp_ops, 0, 0 },
-	{ "U", "udp", 0, "Use UDP datagram (default) for tracerouting",
-			set_ops, &udp_ops, 0, CLIF_EXTRA },
+			set_module, "tcp", 0, 0 },
 	{ "i", "interface", "device", "Specify a network interface "
 			    "to operate with",
 			CLIF_set_string, &device, 0, 0 },
@@ -380,14 +430,16 @@ static CLIF_option option_list[] = {
 			CLIF_set_uint, &sim_probes, 0, 0 },
 	{ "n", 0, 0, "Do not resolve IP addresses to their domain names",
 			CLIF_set_flag, &noresolve, 0, 0 },
-	{ "p", "port", "port", "Use destination port %s. "
-			    "It is an initial value for the UDP destination "
-			    "port (incremented by each probe, default is "
-			    _TEXT(DEF_UDP_PORT) "), for the ICMP "
-			    "seq number (incremented as well, default from 1), "
-			    "and the constant destination port for TCP tries "
-			    "(default is " _TEXT(DEF_TCP_PORT) ")",
-			    CLIF_set_uint, &dst_port_seq, 0, 0 },
+	{ "p", "port", "port", "Set the destination port to use. "
+			    "It is either initial udp port value for "
+			    "\"default\" method (incremented by each probe, "
+			    "default is " _TEXT(DEF_START_PORT) "), "
+			    "or initial seq for \"icmp\" (incremented as well, "
+			    "default from 1), or some constant destination port"
+			    " for other methods (with default of "
+			    _TEXT(DEF_TCP_PORT) " for \"tcp\", "
+			    _TEXT(DEF_UDP_PORT) " for \"udp\", etc.)",
+			    set_port, &dst_port_seq, 0, 0 },
 	{ "t", "tos", "tos", "Set the TOS (IPv4 type of service) or TC "
 			    "(IPv6 traffic class) value for outgoing packets",
 			    CLIF_set_uint, &tos, 0, 0 },
@@ -416,6 +468,28 @@ static CLIF_option option_list[] = {
 			    "registries and print results directly after "
 			    "the corresponding addresses",
 			    CLIF_set_flag, &as_lookups, 0, 0 },
+	{ "M", "module", "name", "Use specified module (either builtin or "
+			    "external) for traceroute operations. Most methods "
+			    "have their shortcuts (`-I' mean `-M icmp' etc.)",
+			    CLIF_set_string, &module, 0, CLIF_EXTRA },
+	{ "O", "options", "OPTS", "Use module-specific option %s for the "
+			    "traceroute module. Several %s allowed, separated "
+			    "by comma. If %s is \"help\", print info about "
+			    "available options",
+			    set_mod_option, 0, 0, CLIF_SEVERAL | CLIF_EXTRA },
+	{ 0, "sport", "num", "Use source port %s for outgoing packets. "
+			    "Implies `-N 1'",
+			    set_port, &src_port, 0, CLIF_EXTRA },
+	{ "U", "udp", 0, "Use UDP to particular port for tracerouting "
+			    "(instead of increasing the port per each probe), "
+			    "default port is " _TEXT(DEF_UDP_PORT),
+			    set_module, "udp", 0, CLIF_EXTRA },
+	{ 0, "UL", 0, "Use UDPLITE for tracerouting (default dest port is "
+			    _TEXT(DEF_UDP_PORT) ")",
+			    set_module, "udplite", 0, CLIF_ONEDASH|CLIF_EXTRA },
+	{ "P", "protocol", "prot", "Use raw packet of protocol %s "
+			    "for tracerouting", 
+			    set_raw, 0, 0, CLIF_EXTRA },
 	CLIF_VERSION_OPTION (version_string),
 	CLIF_HELP_OPTION,
 	CLIF_END_OPTION
@@ -447,7 +521,10 @@ int main (int argc, char *argv[]) {
 	)  exit (2);
 
 
-	if (geteuid () != 0 && ops != &udp_ops)
+	ops = tr_get_module (module);
+	if (!ops)  ex_error ("Unknown traceroute module %s", module);
+
+	if (!ops->user && geteuid () != 0)
 	    ex_error ("The specified type of tracerouting "
 			"is allowed for superuser only");
 
@@ -475,6 +552,14 @@ int main (int argc, char *argv[]) {
 		dst_addr.sin6.sin6_flowinfo =
 			((tos & 0xff) << 20) | (flow_label & 0x000fffff);
 
+	if (src_port) {
+	    src_addr.sin.sin_port = htons ((u_int16_t) src_port);
+	    src_addr.sa.sa_family = af;
+	}
+
+	if (src_port || ops->one_per_time)
+		sim_probes = 1;
+
 
 	/*  make sure we don't std{in|,out,err} to open sockets  */
 	make_fd_used (0);
@@ -489,6 +574,13 @@ int main (int argc, char *argv[]) {
 	probes = calloc (num_probes, sizeof (*probes));
 	if (!probes)  error ("calloc");
 
+
+	if (ops->options && opts_idx > 1) {
+	    opts[0] = strdup (module);	    /*  aka argv[0] ...  */
+	    if (CLIF_parse (opts_idx, opts, ops->options, 0, CLIF_KEYWORD) < 0)
+		    exit (2);
+	}
+
 	if (ops->init (dst_name, &dst_addr, dst_port_seq, packet_len) < 0)
 		ex_error ("trace method's init failed");
 
@@ -499,95 +591,8 @@ int main (int argc, char *argv[]) {
 }
 
 
-/*	POLL  STUFF	    */
-
-static struct pollfd *pfd = NULL;
-static unsigned int num_polls = 0;
-
-void add_poll (int fd, int events) {
-	int i;
-
-	for (i = 0; i < num_polls && pfd[i].fd > 0; i++) ;
-
-	if (i == num_polls) {
-	    pfd = realloc (pfd, ++num_polls * sizeof (*pfd));
-	    if (!pfd)  error ("realloc");
-	}
-
-	pfd[i].fd = fd;
-	pfd[i].events = events;
-}
-
-void del_poll (int fd) {
-	int i;
-
-	for (i = 0; i < num_polls && pfd[i].fd != fd; i++) ;
-
-	if (i < num_polls)  pfd[i].fd = -1;    /*  or just zero it...  */
-}
-
-static int cleanup_polls (void) {
-	int i;
-
-	for (i = 0; i < num_polls && pfd[i].fd > 0; i++) ;
-
-	if (i < num_polls) {	/*  a hole have found   */
-	    int j;
-
-	    for (j = i + 1; j < num_polls; j++) {
-		if (pfd[j].fd > 0) {
-		    pfd[i++] = pfd[j];
-		    pfd[j].fd = -1;
-		}
-	    }
-	}
-
-	return i;
-}
-
-static void do_poll (double timeout) {
-	int nfds, n, i;
-
-	nfds = cleanup_polls ();
-
-	if (!nfds)  return;
-
-	n = poll (pfd, nfds, timeout * 1000);
-	if (n < 0) {
-	    if (errno == EINTR)  return;
-	    error ("poll");
-	}
-
-	for (i = 0; n && i < num_polls; i++) {
-	    if (pfd[i].revents) {
-		ops->recv_probe (pfd[i].fd, pfd[i].revents, probes, num_probes);
-		n--;
-	    }
-	}
-
-	return;
-}
-
-
 /*	PRINT  STUFF	    */
 
-static int equal_addr (const sockaddr_any *a, const sockaddr_any *b) {
-
-	if (!a->sa.sa_family)
-		return 0;
-
-	if (a->sa.sa_family != b->sa.sa_family)
-		return 0;
-
-	if (a->sa.sa_family == AF_INET6)
-	    return  !memcmp (&a->sin6.sin6_addr, &b->sin6.sin6_addr,
-						sizeof (a->sin6.sin6_addr));
-	else
-	    return  !memcmp (&a->sin.sin_addr, &b->sin.sin_addr,
-						sizeof (a->sin.sin_addr));
-	return 0;	/*  not reached   */
-}
-
 static void print_header (void) {
 
 	/*  Note, without ending new-line!  */
@@ -799,8 +804,14 @@ replace_by_final:
 }
 
 
+static void poll_callback (int fd, int revents) {
+
+	ops->recv_probe (fd, revents, probes, num_probes);
+}
+
+
 static void do_it (void) {
-	int start = first_hop * probes_per_hop;
+	int start = (first_hop - 1) * probes_per_hop;
 	int end = num_probes;
 	double last_send = 0;
 
@@ -868,7 +879,7 @@ static void do_it (void) {
 
 		if (timeout < 0)  timeout = 0;
 
-		do_poll (timeout);
+		do_poll (timeout, poll_callback);
 	    }
 
 	}
@@ -904,16 +915,7 @@ void tune_socket (int sk) {
 	}
 
 
-	if (device) {
-	    if (setsockopt (sk, SOL_SOCKET, SO_BINDTODEVICE,
-					device, strlen (device) + 1) < 0
-	    )  error ("setsockopt SO_BINDTODEVICE");
-	}
-
-	if (src_addr.sa.sa_family) {
-	    if (bind (sk, &src_addr.sa, sizeof (src_addr)) < 0)
-		    error ("bind");
-	}
+	bind_socket (sk);
 
 
 	if (af == AF_INET) {
@@ -977,9 +979,9 @@ void tune_socket (int sk) {
 }
 
 
-void parse_icmp_res (probe *pb, int type, int code) {
+void parse_icmp_res (probe *pb, int type, int code, int info) {
 	char *str = "";
-	char buf[16];
+	char buf[sizeof (pb->err_str)];
 
 	if (af == AF_INET) {
 
@@ -1019,7 +1021,8 @@ void parse_icmp_res (probe *pb, int type, int code) {
 			    break;
 
 		    case ICMP_UNREACH_NEEDFRAG:
-			    str = "!F";
+			    snprintf (buf, sizeof (buf), "!F-%d", info);
+			    str = buf;
 			    break;
 
 		    case ICMP_UNREACH_SRCFAIL:
@@ -1076,8 +1079,13 @@ void parse_icmp_res (probe *pb, int type, int code) {
 			    break;
 		}
 	    }
+	    else if (type == ICMP6_PACKET_TOO_BIG) {
+		snprintf (buf, sizeof (buf), "!F-%d", info);
+		str = buf;
+	    }
 	}
 
+
 	if (str && !*str) {
 	    snprintf (buf, sizeof (buf), "!<%u-%u>", type, code);
 	    str = buf;
@@ -1094,6 +1102,102 @@ void parse_icmp_res (probe *pb, int type, int code) {
 }
 
 
+void parse_cmsg (probe *pb, struct msghdr *msg) {
+	struct cmsghdr *cm;
+
+	for (cm = CMSG_FIRSTHDR (msg); cm; cm = CMSG_NXTHDR (msg, cm)) {
+
+	    if (cm->cmsg_level == SOL_SOCKET) {
+
+		if (cm->cmsg_type == SO_TIMESTAMP) {
+		    struct timeval *tv = (struct timeval *) CMSG_DATA (cm);
+
+		    pb->recv_time = tv->tv_sec + tv->tv_usec / 1000000.;
+		}
+	    }
+	    else if (cm->cmsg_level == SOL_IP) {
+
+		if (cm->cmsg_type == IP_TTL)
+			pb->recv_ttl = *((int *) CMSG_DATA (cm));
+		else if (cm->cmsg_type == IP_RECVERR) {
+		    struct sock_extended_err *ee;
+
+		    ee = (struct sock_extended_err *) CMSG_DATA (cm);
+
+		    if (ee->ee_origin == SO_EE_ORIGIN_ICMP) {
+			memcpy (&pb->res, SO_EE_OFFENDER (ee), sizeof(pb->res));
+			parse_icmp_res (pb, ee->ee_type, ee->ee_code,
+								ee->ee_info);
+		    }
+		}
+	    }
+	    else if (cm->cmsg_level == SOL_IPV6) {
+
+		if (cm->cmsg_type == IPV6_HOPLIMIT)
+			pb->recv_ttl = *((int *) CMSG_DATA (cm));
+		else if (cm->cmsg_type == IPV6_RECVERR) {
+		    struct sock_extended_err *ee;
+
+		    ee = (struct sock_extended_err *) CMSG_DATA (cm);
+
+		    if (ee->ee_origin == SO_EE_ORIGIN_ICMP6) {
+			memcpy (&pb->res, SO_EE_OFFENDER (ee), sizeof(pb->res));
+			parse_icmp_res (pb, ee->ee_type, ee->ee_code,
+								ee->ee_info);
+		    }
+		}
+	    }
+	}
+
+
+	if (!pb->recv_time)
+	    pb->recv_time = get_time ();
+
+	return;
+}
+
+
+int equal_addr (const sockaddr_any *a, const sockaddr_any *b) {
+
+	if (!a->sa.sa_family)
+		return 0;
+
+	if (a->sa.sa_family != b->sa.sa_family)
+		return 0;
+
+	if (a->sa.sa_family == AF_INET6)
+	    return  !memcmp (&a->sin6.sin6_addr, &b->sin6.sin6_addr,
+						sizeof (a->sin6.sin6_addr));
+	else
+	    return  !memcmp (&a->sin.sin_addr, &b->sin.sin_addr,
+						sizeof (a->sin.sin_addr));
+	return 0;	/*  not reached   */
+}
+
+
+void bind_socket (int sk) {
+	sockaddr_any *addr, tmp;
+
+	if (device) {
+	    if (setsockopt (sk, SOL_SOCKET, SO_BINDTODEVICE,
+					device, strlen (device) + 1) < 0
+	    )  error ("setsockopt SO_BINDTODEVICE");
+	}
+
+	if (!src_addr.sa.sa_family) {
+	    memset (&tmp, 0, sizeof (tmp));
+	    tmp.sa.sa_family = af;
+	    addr = &tmp;
+	} else
+	    addr = &src_addr;
+
+	if (bind (sk, &addr->sa, sizeof (*addr)) < 0)
+		error ("bind");
+
+	return;
+}
+
+
 void use_timestamp (int sk) {
 	int n = 1;
 
@@ -1101,25 +1205,6 @@ void use_timestamp (int sk) {
 	/*  foo on errors...  */
 }
 
-double get_timestamp (struct msghdr *msg) {
-	struct cmsghdr *cm;
-	double timestamp = 0;
-
-	for (cm = CMSG_FIRSTHDR (msg); cm; cm = CMSG_NXTHDR (msg, cm)) {
-
-	    if (cm->cmsg_level == SOL_SOCKET && cm->cmsg_type == SO_TIMESTAMP) {
-		struct timeval *tv = (struct timeval *)  CMSG_DATA (cm);
-
-		timestamp = tv->tv_sec + tv->tv_usec / 1000000.;
-	    }
-	}
-
-	if (!timestamp)
-		timestamp = get_time ();
-
-	return timestamp;
-}
-
 
 void use_recv_ttl (int sk) {
 	int n = 1;
@@ -1128,21 +1213,58 @@ void use_recv_ttl (int sk) {
 		setsockopt (sk, SOL_IP, IP_RECVTTL, &n, sizeof (n));
 	else if (af == AF_INET6)
 		setsockopt (sk, SOL_IPV6, IPV6_RECVHOPLIMIT, &n, sizeof (n));
-
 	/*  foo on errors   */
-	return;
 }
 
-int get_recv_ttl (struct msghdr *msg) {
-	struct cmsghdr *cm;
-	int ttl = 0;
 
-	for (cm = CMSG_FIRSTHDR (msg); cm; cm = CMSG_NXTHDR (msg, cm)) {
+void use_recverr (int sk) {
+	int val = 1;
 
-	    if ((cm->cmsg_level == SOL_IP && cm->cmsg_type == IP_TTL) ||
-		(cm->cmsg_level == SOL_IPV6 && cm->cmsg_type == IPV6_HOPLIMIT)
-	    )  ttl = *((int *) CMSG_DATA (cm));
+	if (af == AF_INET) {
+	    if (setsockopt (sk, SOL_IP, IP_RECVERR, &val, sizeof (val)) < 0)
+		    error ("setsockopt IP_RECVERR");
+	}
+	else if (af == AF_INET6) {
+	    if (setsockopt (sk, SOL_IPV6, IPV6_RECVERR, &val, sizeof (val)) < 0)
+		    error ("setsockopt IPV6_RECVERR");
+	}
+}
+
+
+void set_ttl (int sk, int ttl) {
+
+	if (af == AF_INET) {
+	    if (setsockopt (sk, SOL_IP, IP_TTL, &ttl, sizeof (ttl)) < 0)
+		    error ("setsockopt IP_TTL");
+	}
+	else if (af == AF_INET6) {
+	    if (setsockopt (sk, SOL_IPV6, IPV6_UNICAST_HOPS,
+						&ttl, sizeof (ttl)) < 0
+	    )  error ("setsockopt IPV6_UNICAST_HOPS");
+	}
+}
+
+
+/*  There is a typo in the kernel before 2.6.22.2, which prevents icmp errors
+  to be obtained by MSG_ERRQUEUE for ipv6 connected raw sockets.
+*/
+static int can_connect = -1;
+
+int raw_can_connect (void) {
+
+	if (can_connect < 0) {
+
+	    if (af == AF_INET)
+		    can_connect = 1;
+	    else {	/*  AF_INET6   */
+		struct utsname uts;
+
+		if (uname (&uts) < 0)
+			return 0;
+
+		can_connect = (strcmp (uts.release, "2.6.22.2") >= 0);
+	    }
 	}
 
-	return ttl;
+	return can_connect;
 }
